/**
    Copyright (C) 2001, 2002, 2003  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdio.h>
#include <dlfcn.h>
#include <tcdirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <tcstring.h>
#include <semaphore.h>
#include <sched.h>
#include <tclist.h>
#include <kernel.h>
#include <tc2.h>
#include <tchash.h>

typedef struct tc2_symbol {
    int type;
    void *sym;
} tc2_symbol_t;

#define TC2_TYPE_SYMBOL 0
#define TC2_TYPE_ALIAS  1

tclist_t *module_dirs;
tchash_table_t *avail_syms, *loaded_syms, *modules, *module_names;
tc2_interface **interfaces;
int unload_modules = 1;

static int
tc2_dep_cmp(const void *a, const void *b)
{
    return strcmp((char *)a, (char *)b);
}

static void *
tc2_init_module(void *p)
{
    tc2_module *module = p, *imod;
    tc2_symbol_t *isym;
    int (*module_init)(char *) = NULL;
    char *init_sym;
    int i;

    pthread_mutex_lock(&module->mut);

    /* Get and run 'init'. Using _tc2_get_symbol to get init would
     * cause a deadlock, so do it the hard way.  */
    kprint("module", TC2_PRINT_DEBUG + 2, "Fetching '_tc2_init' in '%s'.\n",
	   module->module->name);
    init_sym = MKSYM(module->module->name, "_tc2_init");
    tchash_find(loaded_syms, init_sym, -1, &isym);
    free(init_sym);
    if(isym && isym->type == TC2_TYPE_SYMBOL){
	module_init = isym->sym;
	kprint("module", TC2_PRINT_DEBUG + 3, "Got %s:init = %p\n",
	       module->module->name, module_init);
	if(module_init(module->option)){
	    kprint("module", TC2_PRINT_ERROR, "Error in %s:_tc2_init\n",
		   module->module->name);
	    goto out;
	}
    } else {
	kprint("module", TC2_PRINT_ERROR,
	       "Error in module '%s': no '_tc2_init'!\n",
	       module->module->name);
	goto out;
    }

    kprint("module", TC2_PRINT_DEBUG + 2, "Back from %s:init\n",
	   module->module->name);

    for(i = 0; module->interfaces[i]; i++)
	if(!tchash_find(modules, module->interfaces[i]->name, -1, &imod) &&
	   imod == module)
	    module->interfaces[i]->loaded = module;

    module->loaded = 1;

out:
    pthread_cond_broadcast(&module->cond);
    pthread_mutex_unlock(&module->mut);

    return NULL;
}


/* Loads a module */
extern int
tc2_load_module(tc2_module *module, char *opt)
{
    int i = 0;
    tc2_module_info *mi;

    if (module->handle != NULL)
	return 0;                /* Already loaded */

    kprint("module", TC2_PRINT_VERBOSE + 2,
	   "Loading module '%s'.\n", module->module->name);

    /* Open module */
    module->handle = dlopen(module->filename, RTLD_NOW);
    if(!module->handle){
	fputs(dlerror(), stderr);
	return -1;
    }

    for(i = 0; module->module->exports[i].xname != NULL; i++){
	tc2_module *imod;
	tc2_symbol_t *symt;
	char *sym;

	kprint("module", TC2_PRINT_DEBUG + 3, "Exported symbol: %s:%s -> %s\n",
	       module->module->exports[i].interface,
	       module->module->exports[i].xname,
	       module->module->exports[i].iname);

	if(!tchash_find(modules, module->module->exports[i].interface,
			-1, &imod) && imod != module)
	    continue;

	sym = MKSYM(module->module->exports[i].interface,
		    module->module->exports[i].xname);
	symt = malloc(sizeof(*symt));
	if(module->module->exports[i].iname[0] == '@'){
	    symt->type = TC2_TYPE_ALIAS;
	    symt->sym = module->module->exports[i].iname + 1;
	} else {
	    symt->type = TC2_TYPE_SYMBOL;
	    symt->sym =
		dlsym(module->handle, module->module->exports[i].iname);
	    if(!symt->sym){
		kprint("module", TC2_PRINT_ERROR,
		       "Error fetching symbol %s: %s\n",
		       module->module->exports[i].iname, dlerror());
		goto err;
	    }
	}
	if(tchash_replace(loaded_syms, sym, -1, symt, NULL) < 0){
	    kprint("module", TC2_PRINT_ERROR,
		   "Error adding: %s -> %p\n", sym, symt->sym);
	    goto err;
	}

	kprint("module", TC2_PRINT_DEBUG + 4,
	       "Imported symbol: %s -> %p\n", sym, symt->sym);
	free(sym);
    }

    module->autoclean = 0;

    if(!(mi = dlsym(module->handle, "tc2__module_info")) &&
       !(mi = dlsym(module->handle, "tc2__program_info"))){
	kprint("module", TC2_PRINT_ERROR,
	       "Error: missing tc2__module_info.\n");
	goto err;
    }
    mi->this_module = module;

    module->option = opt;

    /* Run module initialization in separate thread. */
    kprint("module", TC2_PRINT_DEBUG + 3, "Starting init thread for %s\n",
	   module->module->name);
    if(pthread_create(&module->init_thread, NULL, tc2_init_module, module)){
	kprint("module", TC2_PRINT_ERROR,
	       "Error starting init thread for %s\n",
	       module->module->name);
	return -1;
    }

    module_count++;

    tc2_call_hook("load_module", module);

    return 0;

err:
    dlclose(module->handle);
    module->handle = NULL;
    return -1;
}

extern int
tc2_del_dep(tc2_module *depender, tc2_module *mod)
{
    if(mod->loaded){
	kprint("deps", TC2_PRINT_DEBUG + 2,
	       "removing dependency %s -> %s\n",
	       mod->module->name, depender->module->name);
	tclist_delete(mod->dep_list, depender->module->name,
		      tc2_dep_cmp, NULL);
    }
    return 0;
}

/* Removes dependency info from dependecy lists */
extern int
tc2_del_all_deps(tc2_module *mod){
    int i, n;
    char **keys;
    tc2_module **modules;

    keys = (char **) tchash_keys(module_names, &n, 1);
    modules = calloc(sizeof(*modules), n);
    for(i = 0; i < n; i++){
	modules[i] = tc2_get_module(keys[i]);
    }
    free(keys);

    for(i = 0; i < n; i++){
	if(modules[i]->loaded){
	    int d = tclist_delete(modules[i]->dep_list,
				mod->module->name, tc2_dep_cmp, NULL);
	    if(!d)
		kprint("deps", TC2_PRINT_DEBUG + 5,
		       "removed dependency %s -> %s\n",
		       modules[i]->module->name, mod->module->name);
	}
    }
    free(modules);

    return 0;
}

extern int
tc2_autoclean(void)
{
    int i, n;
    char **keys;
    tc2_module **modules;

    keys = (char **) tchash_keys(module_names, &n, 1);
    modules = calloc(sizeof(*modules), n);
    for(i = 0; i < n; i++){
	modules[i] = tc2_get_module(keys[i]);
    }
    free(keys);

    for(i = 0; i < n; i++){
	if(modules[i]->loaded){
	    kprint("autoclean", TC2_PRINT_DEBUG + 2, "%s %ld %d\n",
		   modules[i]->module->name,
		   tclist_items(modules[i]->dep_list),
		   modules[i]->autoclean);
	    if((tclist_items(modules[i]->dep_list) == 0) &&
	       modules[i]->autoclean){
		kprint("autoclean", TC2_PRINT_VERBOSE + 2,
		       "Unloading module %s\n", modules[i]->module->name);
		tc2_unload_module(modules[i]);
	    }
	}
    }
    free(modules);

    return 0;
}

/* Unloads a module */
extern int
tc2_unload_module(tc2_module *module)
{
    int (*module_shdn)(void);
    tc2_symbol_t *sym;
    char *shdn_sym;
    int i;

    if(module->handle && module->init_thread){
	pthread_join(module->init_thread, NULL);
	module->init_thread = 0;
    }

    if(module->loaded == 0)
	return 0;

    /* Check dependency table */
    if(tclist_items(module->dep_list) != 0) {
	module->autoclean = 1;
	kprint("module", TC2_PRINT_VERBOSE + 2,
	       "%s: Depending modules found for: %s\n",
	       __FUNCTION__, module->module->name);
	return -1;
    }

    kprint("module", TC2_PRINT_VERBOSE + 2,
	   "Unloading module '%s'.\n", module->module->name);

    tc2_call_hook("unload_module", module);

    /* Get and run 'shutdown' */
    shdn_sym = MKSYM(module->module->name, "_tc2_shutdown");
    tchash_find(loaded_syms, shdn_sym, -1, &sym);
    free(shdn_sym);
    if(sym && sym->type == TC2_TYPE_SYMBOL) {
	module_shdn = sym->sym;
	kprint("module", TC2_PRINT_DEBUG + 3,
	       "%s: running '_tc2_shutdown' in '%s'...\n", __FUNCTION__,
	       module->module->name);
	module_shdn();
	kprint("module", TC2_PRINT_DEBUG + 3,
	       "%s: '_tc2_shutdown' returned.\n", __FUNCTION__);
    } else {
	kprint("module", TC2_PRINT_ERROR,
	       "Error in module '%s': no '_tc2_shutdown'!\n",
		module->module->name);
    }

    /* Remove symbols from hash table. */
    kprint("module", TC2_PRINT_DEBUG + 3,
	   "%s: removing hash table entries for '%s'...\n", __FUNCTION__,
	   module->module->name);
    for(i = 0; module->module->exports[i].xname != NULL; i++) {
	tc2_symbol_t *sp;
	char *sym;
	sym = MKSYM(module->module->name,
		    module->module->exports[i].xname);
	if(!tchash_delete(loaded_syms, sym, -1, &sp))
	    free(sp);
	free(sym);
    }

    /* FIXME: Black magic */
    sched_yield();

    if(unload_modules)
	dlclose(module->handle);
    module->loaded = 0;
    for(i = 0; module->interfaces[i]; i++)
	if(module->interfaces[i]->loaded == module)
	    module->interfaces[i]->loaded = NULL;
    module->handle = NULL;
    module_count--;

    /* Remove dependency */
    kprint("module", TC2_PRINT_DEBUG + 3,
	   "%s: updating dependencies for '%s'...\n", __FUNCTION__,
	   module->module->name);

    _tc2_request(TC2_DEL_ALL_DEPS, NULL, 0,
		 module->module->name);

    /* Remove modules no longer required. */
    _tc2_request(TC2_AUTOCLEAN, NULL, 0);

    return 0;
}

/* Inserts dependency info into dependency tclist_t */
extern int
tc2_add_dep(tc2_module *depender, tc2_module *module)
{
    int s;
    if (depender == module)
	return 0;

    s = tclist_search(module->dep_list, depender->module->name,
		      NULL, tc2_dep_cmp);
    if(s < 0){
	kprint("deps", TC2_PRINT_ERROR,
	       "%s: error creating dependency for: %s\n",
	       __FUNCTION__, depender->module->name);
	return -1;	
    } else if(s > 0){
	kprint("deps", TC2_PRINT_DEBUG + 2,
	       "added dependency %s -> %s\n",
	       module->module->name, depender->module->name);
    }

    return 0;
}

/* Gets a symbol pointer */
static tc2_symbol_t *
get_symbol(tc2_module *cmod, char *interface, char *csym)
{
    tc2_module *module;
    tc2_symbol_t *sym;

    kprint("module", TC2_PRINT_DEBUG + 2, "%s: Getting symbol: %s\n",
	   cmod->module->name, csym);
 
    if(tchash_find(avail_syms, csym, -1, &module) != 0){
	kprint("module", TC2_PRINT_ERROR, "Unresolved symbol: %s\n", csym);
	return NULL;
    }

    if(module->handle == NULL){
	_tc2_request(TC2_LOAD_MODULE, cmod, 1, interface, NULL);
	module->autoclean = 1;
    }

    if(tchash_find(loaded_syms, csym, -1, &sym)){
	kprint("module", TC2_PRINT_ERROR,
	       "Symbol '%s' exists in 'avail_syms', not in 'loaded_syms'\n",
	       csym);
	return NULL;
    }

    if(sym->type == TC2_TYPE_ALIAS){
	if(!(sym = get_symbol(cmod, interface, sym->sym)))
	    return NULL;
    }

    /* Update dependency table */
    if(cmod)
	tc2_add_dep(cmod, module);

    kprint("module", TC2_PRINT_DEBUG + 2, "%s: Got symbol: %s -> %p\n",
	   cmod->module->name, csym, sym->sym);

    return sym;
}

extern void *
_tc2_get_symbol(tc2_module *cmod, char *interface, char *sym)
{
    char *csym = MKSYM(interface, sym);
    tc2_symbol_t *rsym = get_symbol(cmod, interface, csym);
    free(csym);
    return rsym? rsym->sym: NULL;
}

extern tc2_module *
tc2_get_interface(char *interface)
{
    tc2_module *mod;
    if(tchash_find(modules, interface, -1, &mod) != 0){
	kprint("module", TC2_PRINT_VERBOSE + 3,
	       "Can't find interface: %s\n", interface);
	return NULL;
    }
    return mod;
}

extern tc2_module *
tc2_get_module(char *name)
{
    tc2_module *mod;

    if(tchash_find(module_names, name, -1, &mod) != 0){
	kprint("module", TC2_PRINT_ERROR, "Can't find module: %s\n", name);
	return NULL;
    }
    return mod;
}
