/**
    Copyright (C) 2004  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include "kernel.h"
#include <stdio.h>
#include <stdarg.h>
#include <tchash.h>

#define MAX_LEVEL 16

typedef struct tc2_logger {
    char *name;
    tc2_print_t print;
    int valid;
} tc2_logger_t;

typedef struct tc2_logtag {
    tc2_logger_t *levels;
    int size;
} tc2_logtag_t;

static int verbosity;
static tchash_table_t *tags, *loggers;
static int logger_version;
static tc2_logtag_t tag_default;

static int
print_stderr(const char *tag, int level, const char *s, va_list args)
{
    int r = fprintf(stderr, "%s: ", tag);
    r += vfprintf(stderr, s, args);
    return r;
}

extern int
tc2_setprint(char *tag, int start, int end, char *pr)
{
    tc2_logtag_t *log;
    int i;

    if(end > MAX_LEVEL)
	end = MAX_LEVEL;

    if(start > end)
	return -1;

    if(!tag){
	log = &tag_default;
    } else if(tchash_find(tags, tag, -1, &log)){
	log = calloc(1, sizeof(*log));
	tchash_replace(tags, tag, -1, log, NULL);
    }

    if(end >= log->size){
	log->levels = realloc(log->levels, (end + 1) * sizeof(*log->levels));
	log->size = end + 1;
    }

    for(i = start; i <= end; i++){
	log->levels[i].name = pr;
	log->levels[i].valid = 0;
    }

    if(end > verbosity)
	verbosity = end;

    return 0;
}

extern int
tc2_print(char *tag, int lv, const char *fmt, ...)
{
    va_list args;
    tc2_logtag_t *lt;
    int ret;

    if(lv > MAX_LEVEL)
	lv = MAX_LEVEL;

    if(lv > verbosity)
	return 0;

    if(tchash_find(tags, tag, -1, &lt))
	lt = &tag_default;

    if(lt->size <= lv)
	return 0;

    if(!lt->levels[lv].name)
	return 0;

    if(lt->levels[lv].valid < logger_version){
	lt->levels[lv].print = NULL;
	tchash_find(loggers, lt->levels[lv].name, -1, &lt->levels[lv].print);
	lt->levels[lv].valid = logger_version;
    }

    if(!lt->levels[lv].print)
	return 0;

    va_start(args, fmt);
    ret = lt->levels[lv].print(tag, lv, fmt, args);
    va_end(args);

    return ret;
}

extern int
tc2_regprint(char *name, tc2_print_t pr)
{
    tchash_replace(loggers, name, -1, pr, NULL);
    logger_version++;
    return 0;
}

extern int
tc2_delprint(char *name)
{
    tchash_delete(loggers, name, -1, NULL);
    logger_version++;
    return 0;
}

extern int
tc2_init_print(void)
{
    static int init = 0;
    if(init)
	return 0;
    init = 1;

    tags = tchash_new(8, TC_LOCK_SLOPPY, 0);
    loggers = tchash_new(8, TC_LOCK_SLOPPY, 0);

    tc2_regprint("default", print_stderr);
    tc2_regprint("stderr", print_stderr);

    tc2_setprint(NULL, 0, TC2_PRINT_WARNING, "default");

    return 0;
}

static void
free_tag(void *p)
{
    tc2_logtag_t *l = p;
    if(l->levels)
	free(l->levels);
    free(l);
}

extern void
tc2_free_print(void)
{
    tchash_destroy(tags, free_tag);
    tchash_destroy(loggers, NULL);
}

extern void
tc2_verbose(int v)
{
    tc2_init_print();
    tc2_setprint(NULL, verbosity, v, "default");
}

extern void
tc2_debug(int d)
{
    tc2_verbose(TC2_PRINT_DEBUG + d);
}
