/**
    Copyright (C) 2001, 2002, 2003  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdlib.h>
#include <stdio.h>
#include <semaphore.h>
#include <unistd.h>
#include <stdarg.h>
#include <signal.h>
#include <tc2.h>
#include <kernel.h>
#include <tclist.h>

static sem_t req_sem;
static tclist_t *reqs;
static int pdeps;

typedef struct {
    int type;
    int wait;
    sem_t sem_done;
    tc2_module *mod;
    void *data[0];
} kreq;

static void
sigdbg(int s)
{
    pdeps = 1;
    sem_post(&req_sem);
}

extern int
kreq_init(void)
{
    reqs = tclist_new(TC_LOCK_SLOPPY);
    sem_init(&req_sem, 0, 0);
#ifdef DEBUG
    signal(SIGUSR1, sigdbg);
#endif
    return 0;
}

extern void
kreq_free(void)
{
    tclist_destroy(reqs, free);
    sem_destroy(&req_sem);
}

#define sem_value(s) ({ int __v; sem_getvalue(&s, &__v); __v;})

static void
print_deps(void)
{
    int n, i;
    char **keys;

    keys = (char **) tchash_keys(module_names, &n, 1);
    for(i = 0; i < n; i++){
	tc2_module *mod = tc2_get_module(keys[i]);
	if(mod->loaded){
	    tclist_item_t *li = NULL;
	    char *d;
	    fprintf(stderr, "%s%s:", mod->module->name,
		    mod->autoclean? "[ac]": "");
	    while((d = tclist_next(mod->dep_list, &li)))
		fprintf(stderr, " %s", d);
	    fprintf(stderr, "\n");
	}
    }
    free(keys);
}    

extern int
tc2_run(void)
{
    while(module_count > 0 || sem_value(req_sem) > 0){
	kreq *rq;
	tc2_module *mod;

	kprint("request", TC2_PRINT_DEBUG + 3, "waiting for events...\n");
	sem_wait(&req_sem);
	kprint("request", TC2_PRINT_DEBUG + 3, "new event\n");

	if(pdeps){
	    print_deps();
	    pdeps = 0;
	    continue;
	}

	if((rq = tclist_shift(reqs)) == NULL){
	    continue;
	}

	kprint("request", TC2_PRINT_DEBUG + 3,
	       "rq=%p event=%#06x module=%s\n",
	       rq, rq->type, rq->mod? rq->mod->module->name: "(null)");

	switch(rq->type){
	case TC2_LOAD_MODULE:
	    if((mod = tc2_get_interface(rq->data[0])) != NULL ||
	       (mod = tc2_get_module(rq->data[0])) != NULL){
		tc2_load_module(mod, rq->data[1]);
	    }
	    break;

	case TC2_UNLOAD_MODULE:
	    if((mod = tc2_get_interface(rq->data[0])) != NULL ||
	       (mod = tc2_get_module(rq->data[0])) != NULL){
		if(tc2_unload_module(mod) != 0){
		    /* FIXME: Handle error. */
		}
	    }
	    /* FIXME: Handle error. */
	    break;

	case TC2_UNLOAD_ALL:
	{
	    int n,i;
	    char **keys;
	    tc2_module **modules;

	    keys = (char **) tchash_keys(module_names, &n, 1);
	    modules=calloc(sizeof(tc2_module *), n+1);
	    for(i=0; i<n; i++){
		modules[i] = tc2_get_module(keys[i]);
	    }
	    free(keys);

	    for(i = 0; modules[i] != NULL; i++){
		kprint("request", TC2_PRINT_DEBUG + 3, "unloading %s\n",
		       modules[i]->module->name);
		if(tc2_unload_module(modules[i]) != 0){
		    /* FIXME: Handle error. */
		}
	    }
	    free(modules);
	    break;
	}   

	case TC2_AUTOCLEAN:
	    tc2_autoclean();
	    break;

	case TC2_GET_MODULES:
	{
	    int n,i;
	    char **keys;
	    tc2_module **ret;

	    keys = (char **) tchash_keys(module_names, &n, 1);
	    ret=calloc(sizeof(tc2_module *), n+1);
	    for(i=0; i<n; i++){
		ret[i] = tc2_get_module(keys[i]);
	    }
	    free(keys);
	    *(tc2_module ***)rq->data[0] = ret;
	    break;
	}

	case TC2_GET_MODULE:
	    if((mod = tc2_get_module(rq->data[0])) != NULL){
		*(tc2_module **)rq->data[1]=mod;
	    }
	    break;

	case TC2_ABORT:
	    fprintf(stderr,"Kernel aborted.\n");
	    abort();
	    break;

	case TC2_ADD_DEPENDENCY:
	    if((mod = tc2_get_interface(rq->data[0])) != NULL ||
	       (mod = tc2_get_module(rq->data[0])) != NULL)
		tc2_add_dep(rq->mod, mod);
	    break;

	case TC2_DEL_DEPENDENCY:
	    if((mod = tc2_get_interface(rq->data[0])) != NULL ||
	       (mod = tc2_get_module(rq->data[0])) != NULL)
		tc2_del_dep(rq->mod, mod);
	    break;

	case TC2_DEL_ALL_DEPS:
	    if((mod = tc2_get_module(rq->data[0])))
		tc2_del_all_deps(mod);
	    break;

	case TC2_ADD_MODULE:
	    tc2_add_module(rq->data[0], rq->data[1]);
	    break;

	default:
	    kprint("request", TC2_PRINT_ERROR,
		   "***ERROR: %s: unknown event %#06x, rq=%p\n", rq->type, rq);
	    abort();
	}

	kprint("request", TC2_PRINT_DEBUG + 3,
	       "event: %#06x done!\n", rq->type);

	if (rq->wait) {
	    sem_post(&rq->sem_done);
	} else {
	    free(rq);
	}

    }
    return 0;
}

#define rqalloc(n) malloc(sizeof(kreq) + n*sizeof(void *))

extern int
_tc2_request(int type, tc2_module *md, int wait, ...)
{
    kreq *rq = NULL;
    va_list args;

    kprint("request", TC2_PRINT_DEBUG + 3, "event=%#06x module=%s wait=%d\n",
	   type, md? md->module->name: "(null)", wait);

    va_start(args, wait);

    switch(type){
    case TC2_LOAD_MODULE:
	rq = rqalloc(2);
	rq->data[0] = va_arg(args, char *);
	rq->data[1] = va_arg(args, char *);
	break;

    case TC2_UNLOAD_MODULE:
	rq = rqalloc(1);
	rq->data[0] = va_arg(args, char *);
	break;

    case TC2_UNLOAD_ALL:
    case TC2_AUTOCLEAN:
	rq = rqalloc(0);
	break;

    case TC2_GET_MODULES:
	rq = rqalloc(1);
	rq->data[0] = va_arg(args, tc2_module ***);
	break;

    case TC2_GET_MODULE:
	rq = rqalloc(2);
	rq->data[0] = va_arg(args, char *);
	rq->data[1] = va_arg(args, tc2_module **);
	break;

    case TC2_ABORT:
	rq = rqalloc(0);
	break;

    case TC2_ADD_DEPENDENCY:
    case TC2_DEL_DEPENDENCY:
    case TC2_DEL_ALL_DEPS:
	rq = rqalloc(1);
	rq->data[0] = va_arg(args, char *);
	break;

    case TC2_ADD_MODULE:
	rq = rqalloc(2);
	rq->data[0] = va_arg(args, char *);
	rq->data[1] = va_arg(args, tc2_module_info *);
	break;

    default:
	kprint("request", TC2_PRINT_ERROR, "Bad request %i\n", type);
	return -1;
    }

    va_end(args);

    rq->type = type;
    rq->mod = md;
    rq->wait = wait;

    if(wait){
	sem_init(&rq->sem_done, 0, 0);
    }

    kprint("request", TC2_PRINT_DEBUG + 3, "Sending request %#06x rq=%p\n",
	   rq->type, rq);

    tclist_push(reqs, rq);
    sem_post(&req_sem);

    if(wait){
	tc2_module *mod;
	kprint("request", TC2_PRINT_DEBUG + 3,
	       "waiting for request to complete...\n");
	sem_wait(&rq->sem_done);
	kprint("request", TC2_PRINT_DEBUG + 3,
	       "request complete\n");

	/* Wait for module initialization to complete. */
	if(type == TC2_LOAD_MODULE &&
	   (mod = tc2_get_interface(rq->data[0])?:
	    tc2_get_module(rq->data[0]))){
	    kprint("request", TC2_PRINT_DEBUG + 3,
		   "waiting for module %s @ %p to load\n",
		   mod->module->name, mod);
	    pthread_mutex_lock(&mod->mut);
	    while(mod->loaded == 0)
		pthread_cond_wait(&mod->cond, &mod->mut);
	    pthread_mutex_unlock(&mod->mut);
	    kprint("request", TC2_PRINT_DEBUG + 3, "module load completed\n");
	}
	
	sem_destroy(&rq->sem_done);
	free(rq);
    }
    return 0;
}
